'''
Created on Aug 11, 2012

@author: rudd-o
'''

import unittest
try:
    from io import StringIO
except ImportError:
    from io import StringIO
import tempfile
import os
import musictoolbox.synccore as mod
import shutil
from musictoolbox import transcoders

# FIXME: test musictoolbox.synccli

def synthetic_playlists_fixtures():
    p1, p2 = StringIO(), StringIO()
    p1.write('''
    #test playlist
    /home/rudd-o/mp3.mp3
    home/rudd-o/wav.mp3
    ''')
    p2.write('''
    #test playlist 2
    /home/rudd-o/playlist2.mp3
    /home/rudd-o/wav.mp3
    ''')

    p1.seek(0)
    p2.seek(0)

    output = {
              "/home/rudd-o/mp3.mp3":[p1],
              "/home/rudd-o/wav.mp3":[p1, p2],
              "/home/rudd-o/playlist2.mp3":[p2],
              }

    return [p1, p2], output

def mtimes_fixtures():
    t1 = tempfile.NamedTemporaryFile()
    t2 = tempfile.NamedTemporaryFile()
    mt1 = os.stat(t1.name).st_mtime
    mt2 = os.stat(t2.name).st_mtime
    return [t1, t2], [mt1, mt2]

def list_files_recursively_fixtures():
    d = tempfile.mkdtemp()
    files = [os.path.join(d, str(x)) for x in range(10)]
    e = os.path.join(d, "subdir")
    os.mkdir(e)
    giles = [os.path.join(e, str(x)) for x in range(10)]
    for f in files + giles:
        with open(f, "w"):
            pass
    mtimes = [ os.stat(f).st_mtime for f in files + giles ]
    return d, files + giles, mtimes


class DummyTranscoder(transcoders.Transcoder):

    def would_transcode_to(self, from_):
        if from_ == "mp3": return "mp3"
        raise transcoders.CannotTranscode(from_)


class DummyOggToMp3Transcoder(transcoders.Transcoder):

    def would_transcode_to(self, from_):
        if from_ in ["mp3", "ogg"]: return "mp3"
        raise transcoders.CannotTranscode(from_)


class TestParsePlaylists(unittest.TestCase):

    def test_synthetic_playlists(self):
        sources, output = synthetic_playlists_fixtures()
        files, excs = mod.parse_playlists(sources)
        self.assertEqual(files, output)
        self.assertEqual(excs, [])

    def test_nonexistent_playlists(self):
        sources = ["does not exist"]
        files, excs = mod.parse_playlists(sources)
        self.assertEqual(files, {})
        self.assertTrue(excs[0][0], "does not exist")
        self.assertTrue(isinstance(excs[0][1], IOError), excs[0][1])


class TestListFilesRecursively(unittest.TestCase):

    def test_simple_case(self):
        d, files, _ = list_files_recursively_fixtures()
        listed_files = mod.list_files_recursively(d)
        listed_files.sort()
        shutil.rmtree(d)
        self.assertEqual(files, listed_files)


class TestScanMtimes(unittest.TestCase):

    def test_mtimes(self):
        f, t = mtimes_fixtures()
        result = list(mod.scan_mtimes([x.name for x in f]))
        [x.close() for x in f]
        self.assertEqual(result[0][1], t[0])
        self.assertEqual(result[1][1], t[1])


class TestVfatProtect(unittest.TestCase):

    def test_noendingdots(self):
        p = mod.vfatprotect("/some/path/with./dots")
        self.assertNotIn(".", p)
        p = mod.vfatprotect("/some/path/with..../many dots")
        self.assertNotIn(".", p)
        p = mod.vfatprotect("/some/path/with /many spaces")
        self.assertNotIn(" /", p)

    def test_nobadchars(self):
        for f in '?<>\\:*|':
            p = mod.vfatprotect("/path/with%sbadchar" % f)
            self.assertNotIn(f, p)

class TestComputeSynchronization(unittest.TestCase):

    def test_simple_case(self):
        m = mod.compute_synchronization({}, "/", {}, "/")
        self.assertEqual(m, ({}, {}, {}, []))

    def test_identical(self):
        m = mod.compute_synchronization(
            {"/a":1},
            "/",
            {"/target/a":1},
            "/target")
        self.assertEqual(m[0], {})

    def test_older(self):
        m = mod.compute_synchronization(
            {"/a":1},
            "/",
            {"/target/a":0},
            "/target")
        self.assertEqual(m[0], {"/a": "/target/a"})

    def test_absent_target(self):
        m = mod.compute_synchronization(
            {"/a":1},
            "/",
            {},
            "/target")
        self.assertEqual(m[0], {"/a": "/target/a"})

    def test_absent_source(self):
        m = mod.compute_synchronization(
            {},
            "/",
            {},
            "/target")
        self.assertEqual(m[0], {})

    def test_mapper(self):
        mapper = lambda x, y: x.replace("a", "b")
        m = mod.compute_synchronization(
            {"/a":1},
            "/",
            {"/target/b":1},
            "/target",
            path_mapper=mapper)
        self.assertEqual(m[0], {})
        m = mod.compute_synchronization(
            {"/a":1},
            "/",
            {"/target/b":0},
            "/target",
            path_mapper=mapper)
        self.assertEqual(m[0], {"/a": "/target/b"})

    def test_fat32_windowing(self):
        def fatcompare(s, t):
            x = int(s) - int(t)
            if x >= 2: return 1
            elif x <= -2: return -1
            return 0
        m = mod.compute_synchronization(
            {"/a":1},
            "/",
            {"/target/a":0},
            "/target",
            time_comparator=fatcompare)
        self.assertEqual(m[0], {})
        m = mod.compute_synchronization(
            {"/a":2},
            "/",
            {"/target/a":1},
            "/target",
            time_comparator=fatcompare)
        self.assertEqual(m[0], {})
        m = mod.compute_synchronization(
            {"/a":1},
            "/",
            {"/target/a":2},
            "/target",
            time_comparator=fatcompare)
        self.assertEqual(m[0], {})
        m = mod.compute_synchronization(
            {"/a":4},
            "/",
            {"/target/a":2},
            "/target",
            time_comparator=fatcompare)
        self.assertEqual(m[0], {"/a":"/target/a"})

    def test_complex(self):
        s = {"/source/abc":45, "/source/def":30}
        t = {}
        sp = "/"
        tp = "/target"
        e = {
             "/source/abc":"/target/source/abc",
             "/source/def":"/target/source/def",
        }
        m = mod.compute_synchronization(s, sp, t, tp)
        self.assertEqual(m[0], e)

        sp = "/source"
        e = {
             "/source/abc":"/target/abc",
             "/source/def":"/target/def",
        }
        m = mod.compute_synchronization(s, sp, t, tp)
        self.assertEqual(m[0], e)

        sp = "/source/"
        m = mod.compute_synchronization(s, sp, t, tp)
        self.assertEqual(m[0], e)

        tp = "/target/"
        m = mod.compute_synchronization(s, sp, t, tp)
        self.assertEqual(m[0], e)

        sp = "/source"
        m = mod.compute_synchronization(s, sp, t, tp)
        self.assertEqual(m[0], e)

        sp = "/source/a"
        e = {
             "/source/abc":"/target/../abc",
             "/source/def":"/target/../def",
        }
        self.assertRaises(ValueError,
           lambda: mod.compute_synchronization(s, sp, t, tp))


class TestSynchronizer(unittest.TestCase):

    def setUp(self):
        t = DummyTranscoder()
        self.k = mod.Synchronizer(t)

    def test_parse_playlists(self):
        sources, output = synthetic_playlists_fixtures()
        [ self.k.add_playlist(p) for p in sources ]
        self.k._parse_playlists()
        filelist = self.k.source_files
        self.assertTrue(lambda: len(filelist))
        self.assertEqual(self.k.playlists, sources)
        self.assertEqual(self.k.source_files, output)

    def test_scan_source_files_mtimes(self):
        f, t = mtimes_fixtures()
        p1 = StringIO(f[0].name + "\n" + f[1].name)
        p2 = StringIO(f[1].name)
        [ self.k.add_playlist(p) for p in [p1, p2] ]

        try:
            self.k._parse_playlists()
            self.k._scan_source_files_mtimes()
            self.assertTrue(f[0].name in self.k.source_files)
            self.assertTrue(f[1].name in self.k.source_files)
            self.assertTrue(f[0].name in self.k.source_files_mtimes)
            self.assertTrue(f[1].name in self.k.source_files_mtimes)
            self.assertEqual(self.k.source_files_mtimes[f[0].name], t[0])
            self.assertEqual(self.k.source_files_mtimes[f[1].name], t[1])
        finally:
            [x.close() for x in f]

    def test_scan_source_files_mtimes_nofiles(self):
        self.k._parse_playlists()
        self.assertTrue(lambda: len(self.k.source_files) == 0)
        self.assertTrue(lambda: len(self.k.source_files_mtimes) == 0)

    def test_scan_target_dir(self):
        directory, files, _ = list_files_recursively_fixtures()
        self.k.set_target_dir(directory)
        try:
            self.k._scan_target_dir()
            for f in files: self.assertIn(f, self.k.target_files)
        finally:
            shutil.rmtree(directory)

    def test_scan_target_dir_mtimes(self):
        directory, files, mtimes = list_files_recursively_fixtures()
        self.k.set_target_dir(directory)
        try:
            self.k._scan_target_dir()
            self.k._scan_target_dir_mtimes()
            for f in files:
                self.assertIn(f, self.k.target_files)
                self.assertIn(f, self.k.target_files_mtimes)
                expected_mtimes = dict(list(zip(files, mtimes)))
                self.assertEqual(self.k.target_files_mtimes, expected_mtimes)
        finally:
            shutil.rmtree(directory)

    def test_compute_synchronization(self):
        self.k.set_target_dir("/target")
        self.k.scan_done = True  # simulate scanning done
        self.k.source_files_mtimes = {"/source/a.mp3":5}
        self.k.target_files_mtimes = {"/target/a.mp3":1}
        ops = self.k.compute_synchronization()[0]
        self.assertEqual(ops, {"/source/a.mp3": "/target/a.mp3"})

        self.k.target_files_mtimes = {"/target/a.mp3":4}
        self.k.compute_sync_cache = None
        ops = self.k.compute_synchronization()[0]
        self.assertEqual(ops, {})

        self.k.source_files_mtimes = {"/source/a.ogg":7}
        self.k.compute_sync_cache = None
        ops = self.k.compute_synchronization()[0]
        self.assertEqual(ops, {})

        self.k.set_transcoder(DummyOggToMp3Transcoder())
        self.k.scan_done = True  # simulate scanning done
        self.k.compute_sync_cache = None
        ops = self.k.compute_synchronization()[0]
        self.assertEqual(ops, {"/source/a.ogg":"/target/a.mp3"})

        self.k.source_files_mtimes = {"/source/a.MP3":4}
        self.k.target_files_mtimes = {"/target/a.mp3":4}
        self.k.compute_sync_cache = None
        ops = self.k.compute_synchronization()[0]
        self.assertEqual(ops, {})

        self.k.source_files_mtimes = {"/source/a.MP3":7}
        self.k.target_files_mtimes = {"/target/a.mp3":4}
        self.k.compute_sync_cache = None
        ops = self.k.compute_synchronization()[0]
        self.assertEqual(ops, {'/source/a.MP3': '/target/a.mp3'})

        self.k.source_files_mtimes = {"/source/a.MP3":7, "/source/a.mp3":8}
        self.k.target_files_mtimes = {"/target/a.mp3":4, "/target/b.mp3":6}
        self.k.compute_sync_cache = None
        ops = self.k.compute_synchronization()[0]
        self.assertEqual(ops, {'/source/a.MP3': '/target/a.mp3'})
        ops = self.k.compute_synchronization()[2]
        self.assertEqual(ops, {})
        ops = self.k.compute_synchronization()[3]
        self.assertEqual(ops, ["/target/b.mp3"])

        self.k.source_files_mtimes = {"/source/a.MP3":7, "/source/a.mp3":8}
        self.k.target_files_mtimes = {"/target/a.mp3":4, "/target/b.mp3":6}
        self.k.exclude_beneath = ["/target"]
        self.k.compute_sync_cache = None
        ops = self.k.compute_synchronization()[3]
        self.assertEqual(ops, [])
        self.k.exclude_beneath = []

        self.k.source_files_mtimes = {"/source/a.MP3":7}
        self.k.target_files_mtimes = {"/target/a.MP3":4}
        self.k.compute_sync_cache = None
        ops = self.k.compute_synchronization()[0]
        self.assertEqual(ops, {'/source/a.MP3': '/target/a.MP3'})

        self.k.source_files_mtimes = {"/source/a.Ogg":4}
        self.k.target_files_mtimes = {"/target/a.mp3":4}
        self.k.compute_sync_cache = None
        ops = self.k.compute_synchronization()[0]
        self.assertEqual(ops, {})
        ops = self.k.compute_synchronization()[2]  # the skipped files
        self.assertEqual(ops, {"/source/a.Ogg":"/target/a.mp3"})

        self.k.source_files_mtimes = {"/source/a.OgG":9}
        self.k.target_files_mtimes = {"/target/a.mp3":4}
        self.k.compute_sync_cache = None
        ops = self.k.compute_synchronization()[0]
        self.assertEqual(ops, {'/source/a.OgG': '/target/a.mp3'})

        self.k.source_files_mtimes = {"/source/a.OgG":9}
        self.k.target_files_mtimes = {"/target/a.MP3":4}
        self.k.compute_sync_cache = None
        ops = self.k.compute_synchronization()[0]
        self.assertEqual(ops, {'/source/a.OgG': '/target/a.MP3'})

        e = Exception()
        self.k.source_files_mtimes = {"/source/a.OgG":e,
                                    "/source/a.mP3":8}
        self.k.target_files_mtimes = {"/target/a.MP3":4}
        self.k.compute_sync_cache = None
        ops = self.k.compute_synchronization()
        self.assertEqual(ops[0], {'/source/a.mP3': '/target/a.MP3'})
        self.assertEqual(ops[1], {'/source/a.OgG': e})

        self.k.source_files_mtimes = {"/invalidsource/a.OgG":e,
                                    "/source/a.mP3":8}
        self.k.target_files_mtimes = {"/target/a.MP3":4}
        self.k.compute_sync_cache = None
        ops = self.k.compute_synchronization()
        self.assertEqual(ops[0], {'/source/a.mP3': '/target/a.MP3'})
        self.assertEqual(ops[1], {'/invalidsource/a.OgG': e})

    def test_synchronize(self):
        old = mod.SynchronizerSlave._synchronize_wrapper
        def compute_synchronization():
            return {"a":"b"}, {}, {}, []
        def _synchronize_wrapper(self, src, dst):
            return dst
        self.k.compute_synchronization = compute_synchronization
        mod.SynchronizerSlave._synchronize_wrapper = _synchronize_wrapper
        self.k._ensure_directories_exist = lambda _: None
        try:
            sync_tasks = self.k.synchronize()
            number = 0
            for f, result in sync_tasks:
                self.assertEqual(f, "a")
                self.assertEqual(result, "b")
                number += 1
            self.assertEqual(number, 1)
        finally:
            mod.SynchronizerSlave._synchronize_wrapper = old

    def test_synchronize_deletions(self):
        def compute_synchronization(*args, **kwargs):
            return {"/source/a": "/target/a"}, {}, {}, ["/target/pls/zzz"]
        self.k.compute_synchronization = compute_synchronization

        res = []
        self.k.synchronize_deletions(deletor=res.append)
        assert res == ["/target/pls/zzz"]

    def test_bad_synchronize(self):
        self.fired = []
        def compute_synchronization():
            return {"a":"b", "c":"d"}, {}, {}, []
        def transcode(src, dst):
            raise Exception("m")
        self.k.compute_synchronization = compute_synchronization
        self.k.transcoder.transcode = transcode
        self.k._ensure_directories_exist = lambda _: None

        sync_tasks = list(self.k.synchronize())
        results = [ x[1] for x in sync_tasks ]
        assert all(isinstance(x, Exception) for x in results)



import mutagen
import glob
import os
import sys
import logging
import argparse

def sort_by_track_number(paths):
    metadatas = [ (f,mutagen.File(f)) for f in paths ]
    relevant = []
    relevant.extend(
        [ (int(m["tracknumber"][0]),f,m)
           for f,m in metadatas
           if "tracknumber" in m ]
    )
    relevant.extend(
        [ (int(m['TRCK'].text[0].split("/")[0]),f,m)
           for f,m in metadatas
           if "TRCK" in m ]
    )
    relevant.extend(
        [ (99,f,m)
           for f,m in metadatas
           if f not in [ b for a,b,c in relevant] ]
    )
    relevant = list(sorted(relevant))
    return [ f for x,f,w in relevant ]

def make_album_playlist(playlistpath, paths, dryrun=False):
    playlist_dir = os.path.dirname(playlistpath)
    files  = []
    for p in paths:
        if os.path.isdir(p):
            for spec in ['*.mp3', '*.ogg', '*.mpc', '*.flac']:
                files.extend(glob.iglob(os.path.join(p,spec)))
        else:
            files.append(p)
    sortedfiles = sort_by_track_number(files)
    relativefiles = [ os.path.relpath(f, playlist_dir) for f in sortedfiles ]
    if dryrun:
        playlist_fileobj = sys.stdout
    else:
        playlist_fileobj = file(playlistpath, 'wb')
    playlist_fileobj.write("\n".join(relativefiles))

def get_parser():
    program_name = os.path.basename(sys.argv[0])
    program_shortdesc = "Creates a playlist from passed directories and files, sorted by track number"

    parser = argparse.ArgumentParser(
        prog=program_name,
        description=program_shortdesc,
        formatter_class=argparse.RawDescriptionHelpFormatter
    )
    parser.add_argument("-n", "--dry-run", dest="dryrun", action="store_true", help="do nothing except show what will happen [default: %(default)s]")
    parser.add_argument(dest="paths", help="files and directories to include in playlist", nargs='+')
    parser.add_argument(dest="playlist", help="path to an M3U playlist to write", nargs=1)
    return parser

def main(argv=None):
    logging.basicConfig(level=logging.WARNING)

    if argv is None:
        argv = sys.argv
    else:
        sys.argv.extend(argv)

    parser = get_parser()
    args = parser.parse_args()

    make_album_playlist(args.playlist[0], args.paths, dryrun=args.dryrun)
    return 0

if __name__ == "__main__":
    sys.exit(main(sys.argv))



#!/usr/bin/python3

from __future__ import print_function


import sys
import os
import glob
import subprocess
import string
import signal
import logging
import select
import tempfile
import urllib.request, urllib.parse, urllib.error
import urllib.parse
import threading
import queue
import multiprocessing
import time

import mutagen
from mutagen.apev2 import APEv2
from mutagen.id3 import ID3, TXXX
import mutagen.flac
import mutagen.apev2
import mutagen.id3
import mutagen.easyid3
import mutagen.mp3
import mutagen.oggvorbis
import mutagen.musepack




def log(fmt,*args):
	if not args: sys.stderr.write(fmt + "\n")
	else: sys.stderr.write(fmt%args + "\n")
	sys.stderr.flush()

def fixopen(*args,**kwargs):
	# workaround some of my FLAC files have ID3 tags
	# and that causes problems for mutagen
	filename = args[0]
	if filename.lower().endswith(".flac"): return mutagen.flac.Open(*args,**kwargs)
	else:						  return mutagen.File(*args,**kwargs)

def command_available(cmd):
    try: subprocess.call([cmd],stdout=open(os.devnull, "wb"),stderr=open(os.devnull, "wb"))
    except OSError as e:
        if e.errno == 2: return False
        raise
    except subprocess.CalledProcessError as e:
        if e.returncode == 127: return False
    return True

# ======= mp3gain and soundcheck operations ==========

def transform_keys(f,d):
	for k,v in list(d.items()):
		if f(k) != k:
			del d[k]
			d[f(k)] = v
	return d

def convert_gain_from_mp3gain_to_txxx(gain):
	assert type(gain) in (str,str)
	if gain[-3:] == " dB": gain = gain[:-3]
	try: gain = float(gain)
	except ValueError: raise ValueError("invalid gain value %r"%gain)
	return "%.2f dB" % gain

def convert_peak_from_mp3gain_to_txxx(peak):
	assert type(peak) in (str,str)
	try: peak = float(peak)
	except ValueError: raise ValueError("invalid peak value %r"%peak)
	return "%.6f" % peak

def convert_gain_from_txxx_to_mp3gain(gain):
	assert type(gain) in (str,str)
	if gain[-3:] == " dB": gain = gain[:-3]
	try: gain = float(gain)
	except ValueError: raise ValueError("invalid gain value %r"%gain)
	if gain >= 0.0: return "+%f dB" % gain
	return "%f dB" % gain

REPLAYGAIN_TAGS = (
	("replaygain_track_gain", convert_gain_from_mp3gain_to_txxx,convert_gain_from_txxx_to_mp3gain),
	("replaygain_track_peak", convert_peak_from_mp3gain_to_txxx,lambda x: x),
	("replaygain_album_gain", convert_gain_from_mp3gain_to_txxx,convert_gain_from_txxx_to_mp3gain),
	("replaygain_album_peak", convert_peak_from_mp3gain_to_txxx,lambda x: x),
)

def viewmp3norm(files):
	while files:
		file = files.pop(0)

		print() 
		print(file)

		try: tags = ID3(file)
		except Exception as e:
			print("No ID3 tags",e)
			tags = None
		try: apetags = APEv2(file)
		except Exception as e:
			print("No APE tags",e)
			apetags = None

		if apetags:
			print("===APE tags====")
			for key,tag in list(apetags.items()):
				print("%30s"%key,"  ",repr(tag))

		if tags:
			print("===RVA2 tags===")
			for key,tag in list(tags.items()):
				try:
					if tag.desc.lower() in [ x[0] for x in REPLAYGAIN_TAGS ]:
						print("%30s"%key,"  ",repr(tag))
				except AttributeError: continue

def detect_broken_ape_tags(files):
	while files:
		f = files.pop(0)
		try:
			apetags = APEv2(f)
		except mutagen.apev2.APENoHeaderError:
			pass
		except KeyError as e:
			print(f)
		except Exception as e:
			print("while processing %r: %s" % (f, e), file=sys.stderr)


def detect_missing_ape_tags(files):
	while files:
		f = files.pop(0)
		try:
			apetags = APEv2(f)
		except mutagen.apev2.APENoHeaderError:
			print(f)
		except KeyError as e:
			pass
		except Exception as e:
			print("while processing %r: %s" % (f, e), file=sys.stderr)


def viewtags(files):
	while files:
		file = files.pop(0)

		print() 
		print(file)

		try: tags = ID3(file)
		except Exception as e:
			print("No ID3 tags",e)
			tags = None
		try: apetags = APEv2(file)
		except Exception as e:
			print("No APE tags",e)
			apetags = None

		if apetags:
			print("===APE tags====")
			for key,tag in list(apetags.items()):
				print("%30s"%key,"  ",repr(tag))

		if tags:
			print("===ID3 tags===")
			for key,tag in list(tags.items()):
				print("%30s"%key,"  ",repr(tag))


def get_mp3gain_tags(filename):
	"""Helper function
	returns None or a 4-tuple in the order of the above tuple.  some of the values may be None
	"""
	try: i = mutagen.apev2.APEv2(filename)
	except mutagen.apev2.APENoHeaderError: return
	vals = []
	transform_keys(str.lower,i)
	for t in [ x[0] for x in REPLAYGAIN_TAGS ]:
		try: vals.append(i[t])
		except KeyError: vals.append(None)
	def f(s):
		if s: return str(s)
		else: return None
	if any(vals): return [ f(s) for s in vals ]

def get_txxx_tags(filename):
	"""Helper function"""
	try: i = mutagen.id3.ID3(filename)
	except mutagen.id3.ID3NoHeaderError: return
	vals = []
	for t in [ "TXXX:%s"%x[0] for x in REPLAYGAIN_TAGS ]:
		try: vals.append(str(i[t]))
		except KeyError: vals.append(None)
	if any(vals): return [str(s) for s in vals]

def decimalToASCIIHex (number):
	return "%08X"%number

def gain2sc (gain,base):
	result = round((10 ** (-gain / 10.0)) * base);
	if (result > 65534): result = 65534
	return decimalToASCIIHex(result)

def replaygain_to_soundcheck(gain):
	if type(gain) in (str,str): gain = float(gain)
	soundcheck = [ gain2sc(gain, 1000), gain2sc(gain, 1000), gain2sc(gain, 2500), gain2sc(gain, 2500) ]
	soundcheck.append("00024CA8")
	soundcheck.append("00024CA8")
	soundcheck.append("00007FFF")
	soundcheck.append("00007FFF")
	soundcheck.append("00024CA8")
	soundcheck.append("00024CA8")
	return " "+" ".join(soundcheck)

def write_soundcheck_tag(filename,text):
	metadata = mutagen.id3.Open(filename)
	metadata["COMM:iTunNORM:'eng'"] = mutagen.id3.COMM(
			encoding=3,
			lang="eng",
			desc="iTunNORM",
			text=[str(text)])
	metadata.save()

def recalc_soundcheck(filename,mp3gain_values):
	type = config_soundcheck_uses
	assert len(mp3gain_values) == 4
	tg,tp,ag,ap = mp3gain_values
	if type is MP3GAIN_ALBUM:	   	values = [ag,tg]
	elif type is MP3GAIN_TRACK:	values = [tg,ag]
	values = [ v for v in values if v is not None ]
	if not values: return
	value = float(values[0].split()[0])
	debug("Recalculating Soundcheck values")
	tagcontent = replaygain_to_soundcheck(value)
	write_soundcheck_tag(filename,tagcontent)

def mp3gain_to_txxx(filename):
	i = mutagen.apev2.APEv2(filename)
	try: o = mutagen.id3.ID3(filename)
	except mutagen.id3.ID3NoHeaderError: o = mutagen.id3.ID3()
	changed = False
	for name,f,g in REPLAYGAIN_TAGS:
		if name.lower() in i: name = name.lower()
		elif name.upper() in i: name = name.upper()
		else: continue
		o[name] = mutagen.id3.TXXX(desc=name.lower(),encoding=1,text=f(str(i[name])))
		changed = True
	if changed: o.save(filename)

def txxx_to_mp3gain(filename):
	i = mutagen.id3.ID3(filename)
	try: o = mutagen.apev2.APEv2(filename)
	except mutagen.apev2.APENoHeaderError: o = mutagen.apev2.APEv2()
	changed = False
	for name,f,g in REPLAYGAIN_TAGS:
		try:
			o[name.upper()] = g(str(i["TXXX:%s"%name])) ; changed = True
		except KeyError: pass
	if changed: o.save(filename)

def cohere_replaygain_tags(filename):
	debug("Cohering replaygain tags for %r",filename)
	if type(filename) is list:
		for x in filename: cohere_replaygain_tags(x)
		return
	
	mp3gain_tags = get_mp3gain_tags(filename)
	txxx_tags = get_txxx_tags(filename)
	if mp3gain_tags:
		debug("Transforming mp3gain ReplayGain info into TXXX")
		mp3gain_to_txxx(filename) # needs to be done because mp3gain may
							# have altered its values with a volume adjustment, rendering 
							# the content of the TXXX tags invalid
	elif txxx_tags and not mp3gain_tags:
		debug("Transforming TXXX ReplayGain info into mp3gain")
		txxx_to_mp3gain(filename)
	else: return
	mp3gain_values = get_mp3gain_tags(filename)
	recalc_soundcheck(filename,mp3gain_values)

def apply_mp3gain(filename,apply=False):
	debug("Applying replaygain tags for %r with level %s",filename,apply)
	assert apply in (MP3GAIN_ALBUM,MP3GAIN_TRACK,False)
	try: tg,tp,ag,ap = get_mp3gain_tags(filename)
	except TypeError: tg,tp,ag,ap = [None,None,None,None]
	cmd = ["mp3gain","-T","-f"]
	if apply is MP3GAIN_ALBUM:
		if ag: cmd.extend(["-a","-c"])
		else: cmd.extend(["-r","-c"])
	elif apply is MP3GAIN_TRACK: cmd.extend(["-r","-c"])
	else: pass
	cmd = cmd + [filename]
	if len(cmd) == 4: debug("Computing mp3gain ReplayGain with %s"," ".join(cmd))
	else: debug("Applying mp3gain ReplayGain with %s"," ".join(cmd))
	subprocess.check_call(cmd,stdout=open("/dev/null","wb"),stderr=subprocess.STDOUT)
	cohere_replaygain_tags(filename)

# ======= / mp3gain and soundcheck operations ==========


calc_commands = {
	".ogg1"  : "vorbisgain",
	".ogg"   : "vorbisgain -a",
	".flac1" : "metaflac --add-replay-gain",
	".flac"  : "metaflac --add-replay-gain",
	".mp31"  : ["mp3gain -p -t -f",cohere_replaygain_tags],
	".mp3"   : ["mp3gain -p -t -f",cohere_replaygain_tags],
}

undo_commands = {
	".ogg1"  : "vorbisgain -c",
	".ogg"   : "vorbisgain -c",
	".flac1" : "metaflac --remove-replay-gain",
	".flac"  : "metaflac --remove-replay-gain",
	".mp31"  : ["mp3gain -u", "mp3gain -s d"],
	".mp3"   : ["mp3gain -u", "mp3gain -s d"],
}

def mediafilesindir(directory):
	newfiles = []
	for globspec in [ "*%s"%k for k in list(calc_commands.keys()) if not k.endswith("1") ]:
		newfiles.extend( glob.iglob(os.path.join(directory,globspec)) )
	return newfiles

def getalbum(filename):
	metadata = fixopen(filename)
	if not list(metadata.keys()): return None

	try: album = str(metadata["TALB"]).lower()
	except KeyError:
		try: album = str(metadata["album"]).lower() # if keyerror, no tiene album
		except KeyError: album = ''
	# just in case
	if not album.strip(): return None
	return album.strip()

def askdir(directory,album):
	samealbum = set()
	files = mediafilesindir(directory)
	for f in files:
		f = os.path.realpath(f)
		if getalbum(f) == album: samealbum.add(f)
	return samealbum

def askotherfiles(files,album):
	samealbum = set()
	for f in files:
		f = os.path.realpath(f)
		if getalbum(f) == album: samealbum.add(f)
	return samealbum

def alreadyhasreplay(f,album=False):
	try:
		if f.lower().endswith(".mp3"): metadata = mutagen.apev2.Open(f)
		else: metadata = fixopen(f)
	except (mutagen.apev2.APENoHeaderError,KeyError):
		return False
	if album:
		if "replaygain_album_gain" in metadata: return True
		if "REPLAYGAIN_ALBUM_GAIN" in metadata: return True
	else:
		if "replaygain_track_gain" in metadata: return True
		if "REPLAYGAIN_TRACK_GAIN" in metadata: return True
	return False

def alreadyhasrva2(f,album=False):
	try: metadata = mutagen.id3.Open(f)
	except mutagen.id3.ID3NoHeaderError: return False
	for value in list(metadata.values()):
		try:
			if album and value.desc == "replaygain_album_gain": return True
			elif value.desc == "replaygain_track_gain": return True
		except AttributeError: continue
	return False

# algorithm begins here

def doreplaygain(args,options):
	filenames = [ str(os.path.realpath(os.path.abspath(f)),"utf-8") for f in args ]
	for f in filenames[:]:
		if os.path.isdir(f):
			filenames.extend(mediafilesindir(f))
			filenames.remove(f)
	filenames = list(set(filenames))

	while filenames:
		filename = filenames.pop(0)
		if not options.quiet: log( "Doing %s",repr(filename) )
		# initialize our list with it
		files = set([filename])
		album = getalbum(filename)
		if album:
			# add other files from the same directory with the same album
			filesfromdir = askdir(os.path.dirname(filename),album)
			l = len(files)
			files = files | set( [ f for f in filesfromdir ] )
			if not options.quiet: log ( "Found %s new files from the same album in the directory", (len(files)-l) )

			# add other files from the same directory with the same album
			filesfromargs = askotherfiles(filenames,album)
			l = len(files)
			files = files | set( [ f for f in filesfromargs ] )
			if not options.quiet: log ( "Found %s new files from command line arguments", (len(files)-l) )

		# group them by file type :-(
		fileswithexts = [ (os.path.splitext(f)[1].lower(),f) for f in files ]
		groups = {}
		for extension in set([ s[0] for s in fileswithexts ]):
			groups[extension] = [ f[1] for f in fileswithexts if f[0] == extension ]

		for extension,files in list(groups.items()):
			for f in files: assert type(f) is str
			if len(files) == 1:
				album = False
				extension = extension + "1"
			else:
				album = True

			if \
			not options.redo \
			and all( [ alreadyhasreplay(f,album=album) for f in files ] ) \
			and ( "mp3" not in extension or all( [ alreadyhasrva2(f,album=album) for f in files ] ) ):

				if not options.quiet: log( "All (%s) files already have replaygain (album %s), we skip them", len(files),album )

			else:

				# just some verbose code
				withoutape = [ x for x in files if not alreadyhasreplay(x,album=album)]
				if withoutape:
					if not options.quiet: log( "These files do not already have ReplayGain (album %s):"%album)
					for x in withoutape:
						if not options.quiet: log( "     %s",repr(x) )
				if "mp3" in extension:
					withoutrva2 = [ x for x in files if not alreadyhasrva2(x,album=album)]
					if withoutrva2:
						if not options.quiet: log( "These files do not already have RVA2 (album %s):"%album)
						for x in withoutrva2:
							if not options.quiet: log( "     %s",repr(x))
				# / end verbose code
				
				def run_command(command,files):
					if callable(command):
						if options.dryrun: log ( str(command) + "(" + repr(files) + ")" )
						else: command(files)
					else:
						command = command.split()
						command = command + files
						if options.dryrun: command = ["echo"] + command
						subprocess.check_call(command,stderr=subprocess.STDOUT)
				
				if options.redo:
					commands = undo_commands[extension]
					if type(commands) is not list: commands = [commands]
					for command in commands: run_command(command,files)
				
				commands = calc_commands[extension]
				if type(commands) is not list: commands = [commands]
				for command in commands: run_command(command,files)

			for f in files:
				if f in filenames: filenames.remove(f)


# ======= configuration and constants ==========

# Re-encode, even if the source and target formats are the same
# Otherwise, just the following transformations are applied
config_reencode_same_format = False

# Apply MP3 gain process to input and output MP3 files
# Prevents clipping of files with loud volume during decodes
# and also makes every song sound similar in loudness
# Choose whether track or album gain is used, or False if not.
MP3GAIN_TRACK =	1
MP3GAIN_ALBUM =	2
# config_apply_mp3gain = MP3GAIN_ALBUM
# mobile player now supports replaygain
config_apply_mp3gain = False

# Compute replaygain if the file has no replaygain information
# Unfortunately, if replaygain is missing from the transcoded file
# we can only compute the track replaygain, not the album one
# So it's better to computing replaygain on your collection before transcoding
config_compute_missing_replaygain = True

# If any format of ReplayGain information is found in your tracks, then the transcoding process
# will write all other missing formats (TXXX(desc), mp3gain, Soundcheck) during the transcode
# This lets you set the type of ReplayGain that will be used to compute Soundcheck
config_soundcheck_uses = MP3GAIN_ALBUM

# ======= / configuration and constants ==========

def debug(*args): return logging.getLogger("transcoder").debug(*args)
def err(*args): return logging.getLogger("transcoder").exception(*args)
def info(*args): return logging.getLogger().info(*args)
def warning(*args): return logging.getLogger().warn(*args)

# ======= encoding / decoding functions ==========

def maketempcopy(uri):
	"""Create a temporary copy of the contents of the specified URI
	return a path to a temporary file that must be deleted by the caller"""
	extension = os.path.splitext(uri)[1]
	dir = None
	if os.access('/dev/shm',os.W_OK) and not uri.lower().endswith("mpc"): dir = "/dev/shm"
	tmp_fd, tmp_path = tempfile.mkstemp(prefix="transcode-",suffix=extension,dir=dir)
	try:
		tmp_file = os.fdopen(tmp_fd,"w")
		tmp_file.write(urllib.request.urlopen(uri).read(-1))
		tmp_file.flush()
		tmp_file.close()
	except Exception:
		os.unlink(tmp_path)
		raise
	return tmp_path

class TranscoderException(Exception): pass
class NoDecoderException(TranscoderException): pass
class NoEncoderException(TranscoderException): pass

# this shit requires an URGENT rewrite using zope interfaces and adapters, like MP3File needs to be adapted to something FLACFile, and policies to arrange that (may need a private adapter registry then!).

# decoders take the in file and the out file as parameters
decoders = {}
def detect_decoders():
    global decoders
    if command_available("madplay"):
        decoders["mp3"] = lambda i,o: ["madplay","--no-tty-control","--output=wav:%s"%o,i]
    if command_available("ogg123"):
        decoders["ogg"] = lambda i,o: ["ogg123","-d","wav","-f",o,i]
    if command_available("flac"):
        decoders["flac"] = lambda i,o: ["flac","-d","-o",o,i]
    if command_available("mppdec"):
        decoders["mpc"] = lambda i,o: ["mppdec",i,o]
    if command_available("mplayer"):
        decoders["mpc"] = lambda i,o: ["mplayer","-vo","null","-ao","pcm:file=%s"%o,i]
    if command_available("mplayer"):
        decoders['m4a'] = lambda i,o: ["mplayer","-vo","null","-ao","pcm:file=%s"%o,i]
    decoders["wav"] = lambda i,o: ["cp",i,o]
    decoders["flv"] = lambda i,o: ["cp",i,o]
    decoders["mp4"] = lambda i,o: ["cp",i,o]
detect_decoders()

def decode(input,source_format):
	"""receives a path to a file, presumed to be in a temporary directory
	returns a path to a new temporary file that needs to be deleted by the caller"""
	output = input + ".wav"
	decoder = decoders[source_format](input,output)
	debug("Decoding with %s"," ".join(decoder))
	try:	subprocess.check_call(decoder,stdout=open("/dev/null","wb"),stderr=subprocess.STDOUT)
	except Exception:
		if os.path.exists(output): os.unlink(output)
		raise
	return output

# encoders take the in file and the out file as parameters
encoders = {}
def detect_encoders():
    global encoders
    if command_available("lame"):
        encoders["mp3"] = lambda i,o: "lame -q 0 --noreplaygain --nohist --vbr-new".split(" ") + [i,o]
    if command_available("aacplusenc"):
        encoders["aac"] = lambda i,o: ["aacplusenc",i,o,"58"]
    if command_available("ffmpeg"):
        encoders["mp4"] = lambda i,o: ["ffmpeg","-i",i,"-acodec","copy","-vcodec","h264","-vf","scale=800:-1",o]
detect_encoders()

def encode(input,target_format):
	"""receives a path to a file, presumed to be in a temporary directory
	returns a path to a new temporary file that needs to be deleted by the caller"""
	output = input + "." + target_format
	encoder = encoders[target_format](input,output)
	debug("Encoding with %s"," ".join(encoder))
	try:	subprocess.check_call(encoder,stdout=open("/dev/null","wb"),stderr=subprocess.STDOUT)
	except Exception:
		if os.path.exists(output): os.unlink(output)
		raise
	return output

# ======= / encoding / decoding functions ==========

# ======= tag transfer functions ==========

def copy_tag_values(i,o,filter):
	"""copies tag values from i to o, if filter is true"""
	def m(k,v):
		if list(filter(k,v)): o[str(k)] = v
	for k,v in list(i.items()):
		m(k,v)

def copy_generic_tag_values_to_id3(i,o):
	"""copies tag values from i to o, transforming to fit the ID3 spec"""
	tag_transformation_map = {
		"album":"TALB",
		"artist":"TPE1",
		"comment":"COMM",
		"description":"COMM",
		"date":"TDRC",
		"year":"TDRC",
		"title":"TIT2",
		"tracknumber":"TRCK",
		"track":"TRCK",
		"genre":"TCON",
		"replaygain_album_gain":lambda x: mutagen.id3.TXXX(desc="replaygain_album_gain",encoding=1,text=x),
		"replaygain_album_peak":lambda x: mutagen.id3.TXXX(desc="replaygain_album_peak",encoding=1,text=x),
		"replaygain_track_gain":lambda x: mutagen.id3.TXXX(desc="replaygain_track_gain",encoding=1,text=x),
		"replaygain_track_peak":lambda x: mutagen.id3.TXXX(desc="replaygain_track_peak",encoding=1,text=x),
	}
	def iterate(k,v):
		if k in tag_transformation_map:
			if type(v) not in (tuple,list): v = [v]
			constructor = tag_transformation_map[k]
			if callable(constructor):
				newvalues = [ constructor(str(value)) for value in v ]
			else:
				constructor = getattr(mutagen.id3,constructor)
				newvalues = [ constructor(text=str(value),encoding=3) for value in v ]
			list(map(o.add,newvalues))
			return newvalues

	transform_keys(str.lower,i)
	for k,v in list(i.items()): iterate(k,v)

def transfer_tags_any_mp3(origin,destination,tag_reader):

	try: i = tag_reader(origin)
	except Exception:
		err("Could not open source file tag")
		return

	try: o = mutagen.id3.ID3(filename=destination)
	except mutagen.id3.ID3NoHeaderError: o = mutagen.id3.ID3()

	copy_generic_tag_values_to_id3(i,o)
	o.save(filename=destination,v1=2)

def transfer_tags_mp3(origin,destination):

	try:
		ape_tag_found = True
		i = mutagen.apev2.APEv2()
		i.load(origin)
		i.save(destination)
	except mutagen.apev2.APENoHeaderError:
		ape_tag_found = False
	except Exception:
		err("Though it exists, an error blocked opening the APEv2 tag")

	try:
		i = mutagen.id3.ID3()
		i.load(origin)
		i.save(destination,v1=2)
	except mutagen.id3.ID3NoHeaderError:
		pass # no ID3 tag, we skip it
	except Exception:
		err("Though it exists, an error blocked opening the ID3 tag")

tag_transfer_functions = {
	"mp3:mp3": lambda x,y,z,w: transfer_tags_mp3(x,y),
	"mp3:aac": lambda x,y,z,w: transfer_tags_mp3(x,y),
	"ogg:mp3": lambda x,y,z,w: transfer_tags_any_mp3(x,y,mutagen.oggvorbis.Open),
	"flac:mp3": lambda x,y,z,w: transfer_tags_any_mp3(x,y,mutagen.flac.Open),
	"mpc:mp3": lambda x,y,z,w: transfer_tags_any_mp3(x,y,mutagen.musepack.Open),
}

def transfer_tags(origin, destination, source_format=None, target_format=None):
	if not source_format:
		source_format = os.path.splitext(origin)[1][1:].lower()
	if not target_format:
		target_format = os.path.splitext(destination)[1][1:].lower()
	m = "%s:%s"%(source_format,target_format)
	if m in tag_transfer_functions:
		transfer_function = tag_transfer_functions[m]
		transfer_function(origin,destination,source_format,target_format)
	return

# ======= / tag transfer functions ==========

def check_transcodable(source_format,target_format):
        """Gets a string describing the source format, and ( a string describing the
        target format ) or ( a list describing strings of target formats ), then
        chooses the target format that is transcodable to, based on the source
        format, and then returns that.  Raises NoEncoderException or
        NoDecoderException if it is not possible to transcode from the sourrce
        format to any of the target formats."""
    
        global decoders
        global encoders
    
        needs_transcoding = False
        if type(target_format) in (list,tuple):
            if source_format in target_format: needs_transcoding = False
            else: needs_transcoding = True
        else:
            if source_format == target_format: needs_transcoding = False
            else: needs_transcoding = True
        
        if not needs_transcoding:
            return source_format
            
        if source_format not in list(decoders.keys()): raise NoDecoderException(source_format)
        if type(target_format) in (list,tuple):
                if source_format in ["flv","mp4"] and "mp3" in target_format:
                        target_format = list(target_format)
                        target_format.remove("mp3")
                choice = None
                for t in target_format:
                        if t in list(encoders.keys()):
                                choice = t
                                break
                if not choice: raise NoEncoderException(target_format)
                target_format = choice
        else:
                if target_format not in list(encoders.keys()): raise NoEncoderException(target_format)
        return target_format

def transcode(uri,target_format):
	"""receives an URI to the file to transcode, and a three letter format code
	returns a filesystem path to the transcoded file"""
	source_format = uri.split(".")[-1].lower()
	target_format = target_format.lower()

	if source_format.startswith("."): raise Exception("source format cannot start with a dot")
	if target_format.startswith("."): raise Exception("target format cannot start with a dot")
	
	reencode = config_reencode_same_format or "mp4" == target_format # HACK FIXME

	check_transcodable(source_format,target_format)
	
	tempfiles = []
	def delete_later(path):
		tempfiles.append(path)
		return path
	def cleanup():
		print("Deleting the following files: %r"%tempfiles)
		list(map(os.unlink,tempfiles))
		while tempfiles: tempfiles.pop()

	try:

		# create the temporary copy
		tmp_path = delete_later(maketempcopy(uri))

		if source_format == "mp3":
			cohere_replaygain_tags(tmp_path)
			if config_compute_missing_replaygain and not get_mp3gain_tags(tmp_path):
				apply_mp3gain(tmp_path)
			if config_apply_mp3gain:
				apply_mp3gain(tmp_path,config_apply_mp3gain)

		if source_format != target_format or reencode:
			# decode
			wav_path = delete_later(decode(tmp_path,source_format))
			# encode
			dest_path = delete_later(encode(wav_path,target_format))
			# transfer tags
			transfer_tags(tmp_path,dest_path,source_format,target_format)
		else:
			dest_path = tmp_path

		if target_format == "mp3" and source_format != "mp3":
			cohere_replaygain_tags(dest_path)
			if config_compute_missing_replaygain and not get_mp3gain_tags(dest_path):
				apply_mp3gain(dest_path)
			if config_apply_mp3gain:
				apply_mp3gain(dest_path,config_apply_mp3gain)

	except:
		err("Exception occurred during transcode of %r, cleaning up"%uri)
		cleanup()
		raise

	tempfiles.remove(dest_path)
	cleanup()

	return dest_path


# this function prevents parallel makedirs from stomping on each other
makedirlock = multiprocessing.Lock()
def locked_makedirs(dstdir):
	with makedirlock:
		os.makedirs(dstdir)
# end locked makedirs


def transcode_file(src,dst):
	debug("Transcoding %r to %r",src,dst)
	destformat = os.path.splitext(dst)[1][1:]
	newsong = transcode(urllib.request.pathname2url(src),destformat)
	rsync(newsong,dst)
	os.unlink(newsong)
	return dst

def cmdline_process(*args):
	format,path,uri = args
	return format,path,transcode(uri,format)

def parallel_transcode(format,paths):
	uris = [ "file://" + urllib.request.pathname2url(os.path.abspath(path)) for path in paths ]

	items = [ (format,path,uri) for path,uri in zip(paths,uris) ]

	oldsighandler = signal.signal(signal.SIGINT, lambda x,y: sys.exit(0))
	pool = multiprocessing.Pool()
	signal.signal(signal.SIGINT, oldsighandler)

	results = [ pool.apply_async(cmdline_process,item) for item in items ]

	for result in results:
		noresult = True
		while noresult:
			try:
				print("Wait...")
				format,path,transcoded = result.get(1)
				noresult = False
			except multiprocessing.TimeoutError:
				pass
		
		print(path)
		print(transcoded)
	pool.close()
	pool.join()

def relativize(listoffiles,commonprefix=None):
	if not commonprefix: commonprefix = os.path.commonprefix(listoffiles)
	return commonprefix, [ os.path.relpath(x,start=commonprefix) for x in listoffiles ]

def vfatprotect(f):
        f = f.replace("./","/")
        for illegal in '?<>\:*|"^': f = f.replace(illegal,"_")
        while "./" in f: f = f.replace("./","/")
        while " /" in f: f = f.replace(" /","/")
        return f

def rsync(src,dst):
	debug("Rsyncing %r to %r",src,dst)
	return subprocess.check_call(['rsync','--modify-window=5','-t',src,dst])

def transfer(src,dst):
	dr = os.path.dirname(dst)
	if not os.path.isdir(dr):
		try: locked_makedirs(dr)
		except OSError as e:
			if e.errno != 17: raise
	try:
		return transcode_file(src,dst)
	except (NoDecoderException,NoEncoderException) as e:
		rsync(src,dst)
		return dst

def transfer_wrapper(x):
	signal.signal(signal.SIGINT,signal.SIG_IGN)
	info("\nDispatching: %r\n->	%r"%(x))
	try:
		return transfer(*x)
	except Exception as e:
		err("Transfer failed: %r %s",x,e)


class SyncManager:
	source_songs = None
	destination_songs = None
	sourcedir = None
	destdir = None
	threads = None
	lock = threading.RLock()
	exceptions = None
	
	def __init__(self,sourcedir,destdir):
		self.sourcedir = os.path.abspath(sourcedir)
		self.destdir = os.path.abspath(destdir)
		self.playlistdir = os.path.join(self.destdir,"Playlists")
		self.source_songs = {}
		self.source_song_dates = {}
		self.destination_songs = set()
		self.destination_song_dates = {}
		self.threads = []
		self.exceptions = []
		self.do_not_transfer = set()
		self.playlists = []
		
	def source_song_found(self,path,date):
		self.lock.acquire()
		newpath = path
		newpath = os.path.realpath(path)
		newpath = os.path.relpath(newpath,start=self.sourcedir)
		try:
			srcfmt = os.path.splitext(newpath)[1][1:]
			srcfmt = srcfmt.lower()
			newpath = os.path.splitext(newpath)[0]+".mp3"
			newformat = check_transcodable(srcfmt,["mp3","mp4"])
			newpath = os.path.splitext(newpath)[0]+"."+newformat
		except (NoEncoderException,NoDecoderException) as e:
			warning("File %r will be transferred as is because of %r",path,e)
		newpath = vfatprotect(newpath)
		self.source_songs[newpath] = path
		self.source_song_dates[newpath] = date
		if newpath in self.destination_songs and date >= self.destination_song_dates[newpath]:
			#info("Not transferring %r because it is already on the destination device"%newpath)
			self.do_not_transfer.add(newpath)
		self.lock.release()
	
	def destination_song_found(self,path,date):
		self.lock.acquire()
		newpath = path
		newpath = os.path.relpath(newpath,start=self.destdir)
		self.destination_songs.add(newpath)
		self.destination_song_dates[newpath] = date
		if newpath in self.source_songs and date <= self.source_song_dates:
			#info("Not transferring %r because it is already on the destination device"%newpath)
			self.do_not_transfer.add(newpath)
		self.lock.release()
			
	def scan_source(self,source):
		source = os.path.abspath(source)
		# sources are playlist m3u files
		def do_scan():
			sourcedir = os.path.dirname(source)
			files = ( os.path.abspath(os.path.join(sourcedir,x.strip())) for x in open(source).readlines() if x.strip() and not x.startswith("#") )
			for f in files:
				try:
					test = open(f,"r")
					test.close()
					del test
				except Exception as e:
					self.lock.acquire()
					print(f)
					self.exceptions.append(e)
					self.lock.release()
					raise
				try:
					self.source_song_found(f,os.stat(f).st_mtime)
				except Exception as e:
					self.lock.acquire()
					self.exceptions.append(e)
					self.lock.release()
					raise

		self.playlists.append(source)
		t = threading.Thread(target=do_scan)
		self.threads.append(t)
		t.start()
	
	def scan_destination(self):
		destination = self.destdir
		# destinations are directories
		def do_scan():
			for base,dirs,files in os.walk(destination):
				if base == self.playlistdir: continue
				for f in files:
					try:
						thefilename = os.path.join(base,f)
						self.destination_song_found(thefilename,os.stat(thefilename).st_mtime)
					except Exception as e:
						self.lock.acquire()
						self.exceptions.append(e)
						self.lock.release()
						raise
		t = threading.Thread(target=do_scan)
		self.threads.append(t)
		t.start()
	
	def join(self):
		for t in self.threads[:]:
			t.join()
			self.threads.remove(t)
		if self.exceptions:
			raise self.exceptions[0]

	def manifest_transfer(self):
		to_transfer = set(self.source_songs.keys()) - self.do_not_transfer
		srcs_and_dests = [
			(self.source_songs[k],os.path.join(self.destdir,k))
			for k in to_transfer ]
		return srcs_and_dests
                
	def transfer_missing_songs(self):
		srcs_and_dests = self.manifest_transfer()
		
		if srcs_and_dests:
			pool = multiprocessing.Pool(8)
			try:
				result = pool.imap_unordered(
						transfer_wrapper,
						srcs_and_dests,
				)
				i = 0
				total = len(srcs_and_dests)
				for r in result:
					i = i + 1
					log("%s of %s done (%s %%).  Last result: %r"%(i,total,i*100/total,r))
				pool.close()
				pool.join()
			except KeyboardInterrupt:
				info('Interrupted.  Waiting for tasks to finish cleanly.')
				pool.terminate()
				sys.exit(1)

	def transfer_playlists(self):
		# map songs in playlists to targets
		reverse_lookup = dict( [ (y,x) for x,y in list(self.source_songs.items()) ] )
		relplaylistpath = os.path.relpath(self.destdir,start=self.playlistdir)
		for srcp in self.playlists:
			srcpdir = os.path.dirname(srcp)
			destp = os.path.join(self.playlistdir,os.path.basename(srcp))
			info("\nRewriting: %r\n->	   %r"%(srcp,destp))
			lines = [ x.strip() for x in open(srcp).readlines() if x.strip() ]
			f = open(destp,"w")
			for line in lines:
				if line.startswith("#"):
					pass
				else:
					path = os.path.abspath(os.path.join(srcpdir,line))
					path = os.path.join(relplaylistpath,reverse_lookup[path])
					line = path
				f.write(line + "\n")
			f.close()
			
	def remove_obsolete_songs(self):
		songsthatneedtobedeleted = set(self.destination_songs) - set(self.source_songs.keys())
		for song in songsthatneedtobedeleted:
			dst = os.path.join(self.destdir,song)
			info("Removing: %r",dst)
			os.unlink(dst)

	def remove_obsolete_playlists(self):
		existingplaylists = glob.glob(os.path.join(self.playlistdir,"*"))
		totransfer_basenames = set ( [ os.path.basename(b) for b in self.playlists ] )
		existing_basenames = set ( [ os.path.basename(b) for b in existingplaylists ] )
		toremove = existing_basenames - totransfer_basenames
		for p in toremove:
			p = os.path.join(self.playlistdir,p)
			info("Removing: %r",p)
			os.unlink(p)


# FIXME provide args for remove and sourcedir too
def sync_playlists(sourceplaylists,synctodir,dryrun=False):
	remove = True
	sm = SyncManager("/var/shared/Entertainment/Music",synctodir)
	for s in sourceplaylists: sm.scan_source(s)
	sm.scan_destination()
	sm.join()
	if dryrun:
		for x,y in sm.manifest_transfer():
			print("Would transfer %r to %r"%(x,y))
	else:
		sm.transfer_missing_songs()
		sm.transfer_playlists()
		if remove:
			sm.remove_obsolete_playlists()
			sm.remove_obsolete_songs()



'''
Transcoders!
'''

import shutil
from musictoolbox import old
import iniparse
from iniparse import INIConfig
import os
import subprocess
try:
    from urllib.request import pathname2url
except ImportError:
    from urllib import pathname2url


class CannotTranscode(Exception):
    def __init__(self, source_format):
        self.source_format = source_format

    def __str__(self):
        return "<CannotTranscode from=%s>" % (
                                             self.source_format,
                                           )

    def __repr__(self):
        return "<CannotTranscode from=%r>" % (
                                             self.source_format,
                                           )


class Transcoder:

    def would_transcode_to(self, from_):
        '''
        from_ is a file extension (without the leading dot)
        representing the source file format.  The extension passed by
        the caller must be lowercased by the caller.
        
        This method must return the target format as a file extension,
        or raise CannotTranscode(sfmt) if it cannot transcode the file
        in question.
        '''
        raise NotImplementedError

    def would_transcode_file_to(self, src):
        '''
        src is an existing path.

        This method must return the target format as a file extension,
        or raise CannotTranscode(sfmt) if it cannot transcode the file
        in question.
        '''
        raise NotImplementedError

    def transcode(self, source_file, destination_file):
        '''Transcode source_file into destination_file.  Destination_file
        will be overwritten.
        
        The return value is None.  This function is pure side effects.
        
        This function blocks while the transcoding is happening.
        It does not return a deferred.
        '''
        raise NotImplementedError


class AbsentMindedTranscoder(Transcoder):
    '''Doesn't do anything.'''

    def would_transcode_to(self, from_):
        return from_

    def transcode(self, src, dst):
        pass


class CopyTranscoder(Transcoder):
    '''Implementation of a transcoder that just copies files blindly.'''

    def would_transcode_to(self, from_):
        return from_

    def transcode(self, src, dst):
        '''Copy source_file into destination_file'''
        shutil.copyfile(src, dst)


# FIXME: this transcoder should at LEAST detect the formats available
# so it wont fail during sync
class LegacyTranscoder(Transcoder):

    def would_transcode_to(self, from_):
        if from_ in "ogg flac mp3 wav mpc": return "mp3"
        if from_ in "mp4 flv": return "mp4"
        raise CannotTranscode(from_)

    def transcode(self, src, dst):
        old.transcode_file(src, dst)


class FlvMp4WebmToMp3Transcoder(Transcoder):
    '''Transcodes from FLV / MP4 to MP3, avoiding retranscoding if possible.'''

    def would_transcode_to(self, from_):
        if from_ in ["flv", "mp4", "webm", "mkv"]: return "mp3"
        raise CannotTranscode(from_)

    def transcode(self, src, dst):
        '''Transcode FLV / MP4 to MP3 file'''
        output = subprocess.check_output(
                                        [
                                         "ffprobe",
                                         src
                                         ],
                                         stderr=subprocess.STDOUT,
                                         text=True,
                                         )
        if "Audio: mp3" in output:
            src = "file://" + pathname2url(src)
            subprocess.check_call(
                [
                 "gst-launch-1.0",
                 "giosrc", "location=%s" % src,
                 "!", "flvdemux",
                 "!", "audio/mpeg",
                 "!", "filesink", "location=%s" % dst,
                 ],
                stdin=None,
                stdout=None,
                stderr=None,
                close_fds=True,
            )
        else:
            src = "file://" + pathname2url(src)
            subprocess.check_call(
                [
                 "gst-launch-1.0",
                 "giosrc", "location=%s" % src,
                 "!", "decodebin",
                 "!", "audioconvert",
                 "!", "lamemp3enc", "encoding-engine-quality=2", "quality=0",
                 "!", "filesink", "location=%s" % dst,
                 ],
                stdin=None,
                stdout=None,
                stderr=None,
                close_fds=True,
            )


class ExtractAudioTranscoder(Transcoder):
    '''Simply extracts audio from videos into a file.'''

    def would_transcode_file_to(self, src):
        '''Transcode MP4 to M4A file'''
        output = subprocess.check_output(
                                        [
                                         "ffprobe",
                                         src
                                         ],
                                         stderr=subprocess.STDOUT,
                                         text=True,
                                         )
        if "Audio: mp3" in output:
            return "mp3"
        elif "Audio: aac" in output:
            return "m4a"
        else:
            _, ext = os.path.splitext(src)
            raise CannotTranscode(ext[1:])

    def transcode(self, src, dst):
        subprocess.check_call(
            [
             "ffmpeg",
             "-i", src,
             "-acodec", "copy",
             "-vn",
             dst,
             ],
            stdin=None,
            stdout=None,
            stderr=None,
            close_fds=True,
        )


class FlvMp4WebmToWavTranscoder(Transcoder):
    '''Transcodes from FLV / MP4 to RIFF WAVE 32 bit float.'''

    def would_transcode_to(self, from_):
        if from_ in ["flv", "mp4", "webm", "mkv"]: return "wav"
        raise CannotTranscode(from_)

    def transcode(self, src, dst):
        '''Transcode FLV / MP4 to RIFF WAVE file'''
        src = "file://" + pathname2url(src)
        subprocess.check_call(
            [
             "gst-launch-1.0",
             "giosrc", "location=%s" % src,
             "!", "decodebin",
             "!", "audioconvert",
             "!", "audio/x-raw,format=F32LE",
             "!", "wavenc",
             "!", "filesink", "location=%s" % dst,
             ],
            stdin=None,
            stdout=None,
            stderr=None,
            close_fds=True,
        )


class AudioToMp3Transcoder(Transcoder):
    '''Transcodes from any audio format to MP3.'''

    def would_transcode_to(self, from_):
        if from_ in ["ogg", "aac", "m4a", "wav", "flac", "mpc"]:
            return "mp3"
        raise CannotTranscode(from_)

    def transcode(self, src, dst):
        '''Transcode audio file to MP3 file'''
        src = "file://" + pathname2url(src)
        subprocess.check_call(
            [
             "gst-launch-1.0",
             "giosrc", "location=%s" % src,
             "!", "decodebin",
             "!", "audioconvert",
             "!", "lamemp3enc", "encoding-engine-quality=2", "quality=0",
             "!", "filesink", "location=%s" % dst,
             ],
            stdin=None,
            stdout=None,
            stderr=None,
            close_fds=True,
        )


class AudioToWavTranscoder(Transcoder):
    '''Transcodes from any audio format to RIFF WAVE 32 bit float.'''

    def would_transcode_to(self, from_):
        if from_ in ["ogg", "aac", "m4a", "mp3", "flac", "mpc"]:
            return "wav"
        raise CannotTranscode(from_)

    def transcode(self, src, dst):
        '''Transcode audio file to MP3 file'''
        src = "file://" + pathname2url(src)
        subprocess.check_call(
            [
             "gst-launch-1.0",
             "giosrc", "location=%s" % src,
             "!", "decodebin",
             "!", "audioconvert",
             "!", "audio/x-raw,format=F32LE",
             "!", "wavenc",
             "!", "filesink", "location=%s" % dst,
             ],
            stdin=None,
            stdout=None,
            stderr=None,
            close_fds=True,
        )


class ConfigurableTranscoder(Transcoder):
    def __init__(self):
        self.cfg = INIConfig(open(os.path.join(os.path.expanduser("~"),'.syncplaylists.ini')))

    def _lookup_transcoder(self, from_):
        to = getattr(self.cfg.transcoding, from_)
        if isinstance(to, iniparse.config.Undefined):
            to = getattr(self.cfg.transcoding, "*")
        if isinstance(to, iniparse.config.Undefined):
            raise CannotTranscode(from_)

        if to == "copy":
            return CopyTranscoder()

        if to == "extractaudio":
            return ExtractAudioTranscoder()

        known_transcoders = [
                             FlvMp4WebmToMp3Transcoder,
                             AudioToMp3Transcoder,
                             FlvMp4WebmToWavTranscoder,
                             AudioToWavTranscoder,
                             ]

        for t in known_transcoders:
            t = t()
            try:
                can_do = t.would_transcode_to(from_)
            except CannotTranscode:
                continue
            if can_do == to:
                break
        return t

    def would_transcode_to(self, from_):
        return self._lookup_transcoder(from_).would_transcode_to(from_)

    def would_transcode_file_to(self, src):
        from_ = os.path.splitext(src)[1][1:].lower()
        return self._lookup_transcoder(from_).would_transcode_file_to(src)

    def transcode(self, src, dst):
        from_ = os.path.splitext(src)[1][1:].lower()
        transcoder = self._lookup_transcoder(from_)
        return transcoder.transcode(src, dst)



from __future__ import print_function


import sys
import os
from argparse import ArgumentParser
from argparse import RawDescriptionHelpFormatter
from musictoolbox.synccore import Synchronizer
from musictoolbox.transcoders import ConfigurableTranscoder as the_transcoder
import logging
import textwrap


logger = logging.getLogger(__name__)

class SynchronizationCLIBackend:

    debug = False

    def __init__(self,
                 transcoder,
                 destpath,
                 playlists,
                 dryrun,
                 concurrency,
                 delete,
                 exclude_beneath,
                 debug):
        self.synchronizer = Synchronizer(transcoder)
        self.synchronizer.set_target_dir(destpath)
        self.synchronizer.set_exclude_beneath(exclude_beneath or [])
        [ self.synchronizer.add_playlist(p) for p in playlists ]
        self.dryrun = dryrun
        self.concurrency = concurrency
        self.delete = delete
        self.debug = debug

    def run(self):
        """Returns:

        0 if all transcoding / synchronization operations completed.
        2 if scanning experienced a problem.
        4 if a transcoding / synchronization failure took place.
        8 if a problem took place while writing playlists.
        Or a combination of the above.

        Raises Exception in an unexpected failure took place.

        If self.debug == True (see __init__), raises the exceptions
        that the transcoding / sync operations saw.
        """
        exitval = 0
        failures = self.synchronizer.scan()
        if failures:
            print("Problems encountered during scanning:\n", file=sys.stderr)
            for s, t in failures:
                print("Could not scan: %r\nBecause: %r\n" % (s, t), file=sys.stderr)
            print("", file=sys.stderr)
            exitval += 2

        if self.dryrun:
            ops, errors, _, deleting = self.synchronizer.compute_synchronization()
            for s, t in list(ops.items()):
                print("Source: %r\nTarget: %r\n" % (s, t))
            for s, t in list(errors.items()):
                print("Not transferring: %r\nBecause: %r\n" % (s, t))
            if self.delete:
                for t in deleting:
                    print("Deleting: %r\n" % (t,))
            written_playlists, playlist_failures = self.synchronizer.synchronize_playlists(dryrun=True)
            for w in written_playlists:
                print("Would write target playlist %r\n" % w)
            for s, t in playlist_failures:
                print("Could not write: %r\nBecause: %r\n" % (s, t),
                      file=sys.stderr)

            if errors:
                exitval += 4
            if playlist_failures:
                exitval += 8

            return exitval

        sync_tasks = self.synchronizer.synchronize(concurrency=self.concurrency)
        errors = None
        for s, t in sync_tasks:
            if isinstance(t, Exception):
                print("Not synced: %r\nBecause: %s\n" % (s, t),
                      file=sys.stderr)
                if self.debug:
                    raise t
                errors = True
            else:
                print("Synced: %r\nTarget: %r\n" % (s, t),
                      file=sys.stderr)
        if errors:
            exitval += 4

        written_playlists, playlist_failures = self.synchronizer.synchronize_playlists()
        for p in written_playlists:
            print("Written: %r\n" % (p, ), file=sys.stderr)
        if playlist_failures:
            print("Problems while writing playlists:\n", file=sys.stderr)
            for s, t in playlist_failures:
                if self.debug:
                    raise t
                print("Could not write: %r\nBecause: %r\n" % (s, t), file=sys.stderr)
            print("", file=sys.stderr)
            exitval += 8

        if self.delete:
            deletion_failures = self.synchronizer.synchronize_deletions()
            if deletion_failures:
                print("Problems while deleting files:\n", file=sys.stderr)
                for s, t in deletion_failures:
                    if self.debug:
                        raise t
                    print("Could not delete: %r\nBecause: %r\n" % (s, t), file=sys.stderr)
                print("", file=sys.stderr)
                exitval += 8

        return exitval


def run_sync(dryrun, destpath, playlists, concurrency, delete, exclude_beneath, debug):
    """Runs sync process.  Returns what SynchronizatoinCLIBackend.run() does."""
    transcoder = the_transcoder()
    w = SynchronizationCLIBackend(transcoder,
                                  destpath,
                                  playlists,
                                  dryrun,
                                  concurrency,
                                  delete,
                                  exclude_beneath,
                                  debug)
    return w.run()


# cmd line stuff

def get_parser():
    program_name = os.path.basename(sys.argv[0])
    program_shortdesc = "Synchronizes music to a directory."
    program_shortdesc += "\n\n" + textwrap.dedent("""
This program takes a number of file lists / playlists in the command line,
and a destination directory, then synchronizes all the songs in the playlists
to the destination directory, with optional modifications to the files and their
names as they are copied to the destination directory, for example preserving
the path structure except for changing characters incompatible with VFAT
volumes to underscores.

For example, suppose you have two playlists `a.m3u` and `b.m3u` (perhaps
generated with the companion `genplaylist` program) that contain the
respective following song lists:

1. /shared/Music/Ace of Base/Everytime it rains.mp3
2. /shared/Music/Scatman John/Scatman.mp3

1. /shared/Music/Dr. Alban/It's my life (12" mix).mp3

If you run `syncplaylists a.m3u b.m3u "/mounteddrive/Travel Music"`, the
resulting directory structure in `/mounteddrive/Travel Music` will look
like this:

1. /mounteddrive/Travel Music/Ace of Base/Everytime it rains.mp3
2. /mounteddrive/Travel Music/Scatman John/Scatman.mp3
3. /mounteddrive/Travel Music/Music/Dr. Alban/It's my life (12_ mix).mp3
4. /mounteddrive/Travel Music/Playlists/a.m3u
5. /mounteddrive/Travel Music/Playlists/b.m3u

with the paths in the brand new playlists `a.m3u` and `b.m3u` transmuted
so that the song paths are relative to the location of the playlists.  This way
your media player can automatically use the playlists you specified to play
music for you.

`syncplaylists` also has the ability to transcode the files to the preferred
formats of your target device as well.  `syncplaylists will read an INI file
named `~/.syncplaylists.ini` to determine how to transcode.  For example,
if you wanted to transcode all non-MP3 files to MP3, you could add the
following configuration to the file:

    [transcoding]
    mp3=copy
    *=mp3

and then `syncplaylists` will do a best-effort attempt to transcode the files
in your playlists to MP3, except for the MP3 files themselves which would just
get copied as per the `copy` action specified in the configuration file.
    """).strip()

    parser = ArgumentParser(
        prog=program_name,
        description=program_shortdesc,
        formatter_class=RawDescriptionHelpFormatter
    )
    parser.add_argument("-n", "--dry-run", dest="dryrun", action="store_true", help="do nothing except show what will happen [default: %(default)s]")
    parser.add_argument("-d", "--delete", dest="delete", action="store_true", help="remove all files that are not mentioned in the playlists [default: %(default)s]")
    parser.add_argument("-D", "--debug", dest="debug", action="store_true", help="raise exceptions instead of reporting them in a succinct manner [default: %(default)s]")
    parser.add_argument("-e", "--exclude", dest="exclude", action="append", help="ignore all files underneath these paths or that match the specific path [default: %(default)s]")
    parser.add_argument("-v", "--verbose", dest="verbose", action="count", help="set verbosity level [default: %(default)s]")
    parser.add_argument("-c", "--concurrency", metavar="NUMPROCS", dest="concurrency", type=int, default=4, help="number of concurrent processes to run [default: %(default)s]")
    parser.add_argument(dest="playlists", help="paths to M3U playlists to synchronize", metavar="playlist", nargs='+')
    parser.add_argument(dest="destpath", help="destination directory", metavar="dir")
    return parser


def main(argv=None):
    '''Command line options.'''

    if argv is None:
        argv = sys.argv
    else:
        sys.argv.extend(argv)

    parser = get_parser()
    args = parser.parse_args()

    if args.verbose:
        logging.basicConfig(level=logging.INFO)
    else:
        logging.basicConfig(level=logging.WARNING)

    return run_sync(
        dryrun=args.dryrun,
        playlists=args.playlists,
        concurrency=args.concurrency,
        destpath=args.destpath,
        delete=args.delete,
        exclude_beneath=args.exclude,
        debug=args.debug,
    )


if __name__ == "__main__":
    sys.exit(main())



'''
Created on Aug 11, 2012

@author: rudd-o
'''

import collections
import logging
import os
import signal as _
from threading import Thread, Lock
from queue import Queue, PriorityQueue
import musictoolbox.old

logger = logging.getLogger(__name__)

# ================== generic utility functions ================

def parse_playlists(sources):
    '''
    Take several playlists, and return a dictionary and a list.
    This dictionary is structured as follows:
    - keys: absolute path names mentioned in the playlists
    - values: playlists where the file appeared
    The list is a sequence of (file, Exception) occurred while parsing.
    '''
    files = {}
    excs = []
    for source in sources:
        try:
            sourcename = source
            if hasattr(source, "read"):
                sourcedir = "/"
            else:
                sourcedir = os.path.dirname(source)
                source = open(source)
            thisbatch = [
                     os.path.abspath(
                         os.path.join(sourcedir, x.strip())
                     )
                     for x in source.readlines()
                     if x.strip() and not x.strip().startswith("#")
            ]
            for path in thisbatch:
                if path not in files: files[path] = []
                files[path].append(sourcename)
        except Exception as e:
            excs.append((source, e))
    return files, excs

def scan_mtimes(filelist):
    '''Take a list of files, and yield tuples (path, mtime or Exception).
    
    The dictionary value for a particular key may contain an
    exception in lieu of an mtime.  This means that scanning
    the key failed.
    '''
    def mtimeorexc(x):
        try:
            return os.stat(x).st_mtime
        except Exception as e:
            return e

    for f in filelist:
        yield f, mtimeorexc(f)

def chunkify(longlist, nchunks):
    return [
          longlist[i::nchunks]
          for i in range(nchunks)
          if longlist[i::nchunks]
    ]

def merge_dicts(d):
    '''take a list of dicts and merge it into a single dict'''
    r = {}
    list(map(r.update, d))
    return r

def get_deferredlist_return_column(x):
    '''take a table of deferredlist's (result, return) outputs and return
    only the return column.'''
    if x: return list(zip(*x))[1]
    return x

def assert_deferredlist_succeeded(result):
    if result:
        assert all(x is True for x in zip(*result)[0]), "deferredlist failed: %r" % result
    return result

def list_files_recursively(directory):
    '''Return a list of absolute paths from recursively listing a directory'''
    return [
            os.path.abspath(os.path.join(base, f))
            for base, _, files in os.walk(directory)
            for f in files
            ]

def vfatprotect(f):
    for illegal in '?<>\\:*|"^': f = f.replace(illegal, "_")
    while "./" in f: f = f.replace("./", "/")
    while " /" in f: f = f.replace(" /", "/")
    return f

class VFATMapper(object):

    def __init__(self, extant_paths, extension_transmogrifier=None):
        """The mapper takes a list of existent paths on the target, rooted
        on the same directory passed to map(path, ...) and an optional extension
        transmogrifier that will change the extension of the file as needed,
        for the case when the file gets transcoded."""
        # extension_transmogrifier takes the original path passed to map()
        # and returns the tentative new extension with the leading dot included
        self.paths_seen = dict(list(zip((x.lower() for x in extant_paths),
                                   extant_paths)))
        self.extension_transmogrifier = extension_transmogrifier

    def map(self, path, originalpath):
        """The map function takes a path rooted on the target directory
        (/target/a is passed as a, as were extant_paths passed to __init__)
        and the full path to originalpath such that the extension
        transmogrifier can open and read the original path if necessary."""
        # the mapper function needs to take FAT32 into account, and needs
        # to know which file formats the targets will be, so it will ask
        # the transcoder about that
        # since FAT32 is case-insensitive but case-preserving, we look
        # up the path in a cache, to select the preferred path
        if self.extension_transmogrifier:
            base, ext = os.path.splitext(path)
            # if the file has an extension
            if ext:
                # check if the extension would change
                newext = self.extension_transmogrifier(originalpath)
                # if something came back from the lookup:
                if newext:
                    # if the extension is the same, just use the original
                    if newext.lower() == ext.lower():
                        newext = ext
                # and reconstitute the dotted extension
                path = base + newext
        path = vfatprotect(path)
        if path.lower() in self.paths_seen:
            path = self.paths_seen[path.lower()]
        return path

def fatcompare(s, t):
    x = int(s) - int(t)
    if x >= 2: return 1
    elif x <= -2: return -1
    return 0

def compute_synchronization(
    sources, sourcedir,
    targets, targetdir,
    path_mapper=lambda x, unused_y: x,
    time_comparator=lambda x, y: 1 if x > y else 0 if x == y else -1,
    unconditional=False,
    exclude_beneath=None,
    ):
    '''
    Compute a synchronization schedule based on a dictionary of
    {source_filename:mtime} and a dictionary of {target_filename:mtime}.
    Paths must be absolute or the behavior of this function is undefined.
    All paths must be contained in their respective specified directories.
    mtime in the source dictionary can be an exception, in which case it will
    be presumed that the source file in question cannot be transferred.

    The synchronization aims to discover the target file names based on
    the source file names, and then automatically figure out which files
    need to be transferred, based on their absence in the remote site,
    or their modification date.

    This function accepts a path mapping callable that will transform
    a partial source file name, and the full path to the source file name,
    into the desired partial target file name prior to performing the
    date comparison.  By default, it is an identity function x, y => x.

    This function also accepts a time comparator function that will
    be used to perform the comparison between source and target
    modification times.  The comparator is the standard 1 0 -1 comparator
    that takes x,y where it returns 1 if x > y, 0 if x and y are identical,
    and -1 if x < y.  The comparator gets passed the source mtime as the
    first parameter, and the target mtime as the second parameter.
    FAT file system users may want to pass a custom comparator that takes
    into consideration the time resolution of FAT32 file systems
    (greater than 2 seconds). 

    Return four values in a tuple:
        1. A dictionary {s:t} where s is the source file name, and
           t is the desired target file name after transfer.
        2. A dictionary {s:e} where s is the source file name, and
           e is the exception explaining why it cannot be synced,
        3. A dictionary {s:e} of files that will be skipped, with their
           corresponding would-be targets that already were transferred,
        4. A list of files that will be deleted from the destination.
    '''
    if exclude_beneath is None:
        exclude_beneath = []
    exclude_beneath = [os.path.abspath(p) for p in exclude_beneath]
    wont_transfer = dict([
        (k, v) for k, v in list(sources.items())
        if isinstance(v, Exception)
    ])
    source_mtimes = dict([
        (k, v) for k, v in list(sources.items())
        if not isinstance(v, Exception)
    ])
    source_files = list(source_mtimes.keys())
    source_basedir = sourcedir
    target_mtimes = targets
    target_files = list(target_mtimes.keys())
    target_basedir = os.path.abspath(targetdir)

    test = source_basedir + os.path.sep \
           if source_basedir[-len(os.path.sep)] != os.path.sep \
           else source_basedir
    for k in source_files:
        if not k.startswith(test):
            raise ValueError(
                "source path %r not within source dir %r" % (k, source_basedir)
            )

    test = target_basedir + os.path.sep \
           if target_basedir[-len(os.path.sep)] != os.path.sep \
           else target_basedir
    for k in target_files:
        if not k.startswith(test):
            raise ValueError(
                "target path %r not within target dir %r" % (k, target_basedir)
            )

    desired_target_files = collections.OrderedDict()
    new_source_files = []
    for p in source_files:
        try:
            mapped_target_file = os.path.join(
                target_basedir, path_mapper(
                    os.path.relpath(p, start=source_basedir), p
                )
            )
            desired_target_files[mapped_target_file] = mapped_target_file
            new_source_files.append(p)
        except Exception as e:
            wont_transfer[p] = e
    source_files = new_source_files

    # desired target to original source map
    dt2s_map = list(zip(source_files, list(desired_target_files.keys())))
    map_of_transfer = [
         (s,
          t,
          t not in target_files
          or time_comparator(source_mtimes[s], target_mtimes[t]) > 0
          or unconditional
          ) for s, t in dt2s_map
    ]
    need_to_transfer = dict([
        (s, t) for s, t, transfer in map_of_transfer
        if transfer and
        not any((t.startswith(x + os.path.sep) or t == x) for x in exclude_beneath)
    ])
    skipping_transfer = dict([
        (s, t) for s, t, transfer in map_of_transfer
        if not transfer and
        not any((t.startswith(x + os.path.sep) or t == x) for x in exclude_beneath)
    ])
    deleting = [
        t for t in target_files
        if t not in desired_target_files and
        not any((t.startswith(x + os.path.sep) or t == x) for x in exclude_beneath)
    ]

    return need_to_transfer, wont_transfer, skipping_transfer, deleting


_dir_lock = Lock()


def ensure_directories_exist(dirs):
    with _dir_lock:
        for t in dirs:
            if not t:
                continue
            if not os.path.exists(t):
                os.makedirs(t)

#==================================================================


#===================== synchronizer code ==========================

class SynchronizerSlave(Thread):
    def __init__(self, inqueue, outqueue, transcoder, postprocessor):
        Thread.__init__(self)
        self.inqueue = inqueue
        self.outqueue = outqueue
        self.transcoder = transcoder
        self.postprocessor = postprocessor

    def run(self):
        while True:
            item = self.inqueue.get()
            unused_prio, (src, dst) = item
            if src is None:
                break
            try:
                self._synchronize_wrapper(src, dst)
                self.outqueue.put((src, dst))
            except BaseException as e:
                self.outqueue.put((src, e))

    def _synchronize_wrapper(self, s, d):
        # adapt the transcoder interface to return the source file
        # because the transcoder returns old and new formats
        target_dir, target_file = os.path.dirname(d), os.path.basename(d)
        tempf = 'tmp-' + target_file
        try:
            ensure_directories_exist([target_dir])
        except BaseException:
            try:
                os.rmdir(target_dir)
            except FileNotFoundError:
                pass
            raise

        tempdest = os.path.join(target_dir, tempf)
        try:
            newext = self.transcoder.transcode(s, tempdest)
            self.postprocessor(s, tempdest, target_format=newext)
        except BaseException:
            try:
                os.unlink(tempdest)
            except FileNotFoundError:
                pass
            raise
        if (os.stat(tempdest).st_size == 0
            and
            os.stat(s).st_size != 0):
            os.unlink(tempdest)
            raise AssertionError("we expected the transcoded file to be "
                                 "larger than 0 bytes")
        if newext is not None:
            dpath, _ = os.path.splitext(d)
            d = dpath + "." + newext
        os.rename(tempdest, d)
        return d

    def stop(self):
        self.stopped = True


class Synchronizer(object):

    playlists = None
    target_dir = None

    source_files = None
    source_files_mtimes = None

    _target_files = None
    def _get_target_files(self):
        return self._target_files
    def _set_target_files(self, target_files):
        self._target_files = target_files
    target_files = property(
                            _get_target_files,
                            _set_target_files
                            )

    _target_files_mtimes = None
    def _get_target_files_mtimes(self):
        return self._target_files_mtimes
    def _set_target_files_mtimes(self, target_files_mtimes):
        # since test code manually sets the files_mtimes without setting
        # the files list, we refresh the files list using the keys
        # of this lookup
        self._target_files_mtimes = target_files_mtimes
        self.target_files = list(target_files_mtimes.keys())
    target_files_mtimes = property(
                                   _get_target_files_mtimes,
                                   _set_target_files_mtimes
                                   )

    def __init__(self, transcoder):
        '''
        Class instance initializer.

        transcoder must be a Transcoder instance that Synchronizer
        can query to determine the target formats available and what
        formats the files will be transcoded to.
        '''
        self.playlists = []
        self.target_dir = None
        self.source_files = {}
        self.source_files_mtimes = {}
        self.target_files = []
        self.target_files_mtimes = {}
        self.exclude_beneath = []
        self.transcoder = transcoder
        self.compute_sync_cache = None
        self.scan_done = False

    @property
    def target_playlist_dir(self):
        return os.path.join(self.target_dir, "Playlists")

    def set_transcoder(self, transcoder):
        '''Change to a different transcoder'''
        self.transcoder = transcoder
        self.scan_done = False

    def set_target_dir(self, target_dir):
        self.target_dir = target_dir
        self.scan_done = False

    def add_playlist(self, playlist):
        self.playlists.append(playlist)
        self.scan_done = False

    def set_exclude_beneath(self, paths):
        self.exclude_beneath = list(paths)
        self.scan_done = False

    def scan(self):
        """Scan all files in added playlists to later compute synchronization.

        Returns a list of (path, Exception)."""
        self.compute_sync_cache = None

        excs = []

        logger.info("Parsing %s playlists", len(self.playlists))
        excs.extend(self._parse_playlists())
        logger.info("Discovered %s source files",
                    len(self.source_files))

        logger.info("Scanning target directory %s", self.target_dir)
        excs.extend(self._scan_target_dir())
        logger.info("Discovered %s target files", len(self.target_files))

        logger.info("Scanning %s source files mtimes",
                    len(self.source_files))
        excs.extend(self._scan_source_files_mtimes())
        logger.info("Scanned %s source files mtimes",
                    len(self.source_files_mtimes))

        logger.info("Scanning %s target files mtimes",
                    len(self.target_files))
        excs.extend(self._scan_target_dir_mtimes())
        logger.info("Scanned %s target files mtimes",
                    len(self.target_files_mtimes))

        self.scan_done = True
        return excs

    def _parse_playlists(self):
        '''
        scan the playlists known to the synchronizer and obtain a list
        of files.

        when done, update self.source_files with the files discovered
        from the playlists, then return exceptions that happened during
        parsing, as a list of tuples (file, exception).

        It is not an error to call this method before having used
        add_playlist(), but it will produce no result.

        This operation blocks.
        '''
        self.source_files, excs = parse_playlists(self.playlists)
        return excs

    def __generic_scan_mtimes_of_file_list(self, filenames):
        '''
        scan files in file_list for their modification
        times. do the scan in parallel for maximum performance.
        Returns dictionary {filename: mtime or Exception}.
        
        This operation blocks.
        '''
        chunks = chunkify(filenames, 8)
        # FIXME TODO hardcoded 8

        class MtimeScanner(Thread):
            def __init__(self, queue, files):
                Thread.__init__(self)
                self.queue = queue
                self.files = files
            def run(self):
                for ret in scan_mtimes(self.files):
                    self.queue.put(ret)
                self.queue.put(None)

        queue = Queue()
        threads = [ MtimeScanner(queue, c) for c in chunks ]
        [ t.start() for t in threads ]
        donecount = 0
        resultdict = {}
        while donecount < len(chunks):
            val = queue.get()
            if val is None:
                donecount += 1
            else:
                resultdict[val[0]] = val[1]
        [ t.join() for t in threads ]
        return resultdict

    def _scan_source_files_mtimes(self):
        '''
        scan all the files in source_files for their modification
        times. do the scan in parallel for maximum performance.
        
        this function must be called after self.parse_playlists()
        
        when done, update self.source_files_mtimes with the modification
        times of all files in self.source_files. 
        
        It is not an error to call this method before having used
        scan_source_files(), but it will produce no result.
        
        This operation blocks.
        '''
        filenames = list(self.source_files.keys())
        files_mtimes = self.__generic_scan_mtimes_of_file_list(filenames)
        self.source_files_mtimes = files_mtimes
        return [ x for x in list(self.source_files_mtimes.items())
                if isinstance(x[1], Exception) ]

    def _scan_target_dir(self):
        '''
        scan the target dir known to the synchronizer.  the scan is
        done in a serialized manner.
        
        when done, update self.target_files with a list of all known files
        in self.target_dir.
        
        It is an error to call this method before set_target_dir()
        
        This operation blocks.
        '''
        try:
            d = list_files_recursively(self.target_dir)
            self.target_files = d
            return []
        except Exception as e:
            return [(self.target_dir, e)]

    def _scan_target_dir_mtimes(self):
        '''
        scan all the files in target_files for their modification
        times. do the scan in parallel for maximum performance.
        
        this function must be called after self.scan_target_dir()
        
        when done, update self.target_files_mtimes with the modification
        times of all files in self.target_files. 
        
        It is not an error to call this method before having used
        scan_target_dir(), but it will produce no result.
        
        This operation blocks.
        '''
        filenames = self.target_files
        files_mtimes = self.__generic_scan_mtimes_of_file_list(filenames)
        self.target_files_mtimes = files_mtimes
        return [ x for x in list(self.target_files_mtimes.items())
                if isinstance(x[1], Exception) ]

    def compute_synchronization(self, unconditional=False):
        '''Computes synchronization between sources and target.

        You must have already parsed the playlists, scanned the target
        directory, and scanned the mtimes of all sources and targets.

        This function does not block.  It returns a pair of dictionaries
        where the first dictionary is a manifest of synchronization operations
        {source:target} that must be executed to bring the target up-to-date
        with respect to the source, the second dictionary is a manifest
        of {source:reason} where source is a source path name and reason
        is an exception explaining why it won't be transferred, the third
        dictionary contains files that should be skipped, and the fourth
        list is a list of files that must be deleted.

        This function has a preference for FAT32 target file systems. In
        an ideal world, it would detect the file system capabilities of
        the target file system.
        '''
        assert self.scan_done, "programming error: must scan first"

        # let's generate the source basedir
        # due to a problem in os.path.commonprefix
        # where two paths /a/bbb/ccc/ddd.mp3
        # and             /a/bbb/cc catch/cadillac.mp3
        # return /a/bbb/cc instead of /a/bbb/
        # we check that the commonprefix is terminated by a slash
        # and if not, we eat up the end of the string until a slash
        # appears, or the string disappears (in which case => /)

        if self.compute_sync_cache:
            return self.compute_sync_cache

        source_basedir = os.path.commonprefix([
            k for k, v in list(self.source_files_mtimes.items())
            if not isinstance(v, Exception)
        ])

        while source_basedir and source_basedir[-1] != os.path.sep:
            source_basedir = source_basedir[:-len(os.path.sep)]
        if not source_basedir: source_basedir = os.path.sep

        # now we filter source files according to whether they can
        # be transcoded or not, since we won't be transferring the ones
        # that cannot be transcoded
        # we lowercase the extension in order to look it up properly
        # because the transcoder may not understand uppercase ones
        source_files_mtimes = dict(list(self.source_files_mtimes.items()))

        def extension_transmogrifier(originalpath):
            _, ext = os.path.splitext(originalpath)
            if ext:
                ext = ext[1:]
                try:
                    newext = self.transcoder.would_transcode_to(ext.lower())
                except NotImplementedError:
                    newext = self.transcoder.would_transcode_file_to(originalpath)
                return "." + newext
            else:
                return ""

        mapper = VFATMapper([os.path.relpath(
                                x,
                                self.target_dir
                            ) for x in list(self.target_files_mtimes.keys())],
                            extension_transmogrifier)

        def comparer(s, t):
            return fatcompare(s, t)

        self.compute_sync_cache = compute_synchronization(
            source_files_mtimes,
            source_basedir,
            self.target_files_mtimes,
            self.target_dir,
            path_mapper=mapper.map,
            time_comparator=comparer,
            unconditional=unconditional,
            exclude_beneath=self.exclude_beneath + [self.target_playlist_dir],
        )
        return self.compute_sync_cache

    def synchronize(self, concurrency=1):
        '''
        Computes synchronization between sources and target, then
        gets ready to sync using a thread pool, dispatching tasks to it.
        The start of the sync process happens immediately.

        The entry conditions for this function are the same as the conditions
        for the self.compute_synchronization function. You must wait until
        a synchronization is done to synchronize again.

        This function blocks.  As it processes files in its plan, it returns
        a tuple (source_file:destination_file or Exception).
        '''
        inqueue = PriorityQueue()
        outqueue = PriorityQueue()
        will_sync, _, _, _ = self.compute_synchronization()
        if not list(will_sync.items()):
            return

        threads = [SynchronizerSlave(inqueue,
                                     outqueue,
                                     self.transcoder,
                                     musictoolbox.old.transfer_tags)
                   for _ in range(concurrency)]
        [ t.start() for t in threads ]
        logger.info("Synchronizing with %s threads for %s work items",
                    len(threads), len(list(will_sync.items())))

        try:
            for s, t in list(will_sync.items()):
                inqueue.put((0, (s, t)))
            for _ in range(concurrency):
                inqueue.put((1, (None, None)))
            outs = 0
            while outs < len(will_sync):
                val = outqueue.get()
                outs = outs + 1
                yield val
        finally:
            for _ in range(concurrency):
                inqueue.put((-1, (None, None)))
            logger.info("Stopping and joining all %s threads", len(threads))
            [ t.join() for t in threads ]
            logger.info("Ended synchronization")

    def synchronize_playlists(self, dryrun=False):
        '''
        Once synchronization of files has been done, synchronization of
        playlists is possible.

        The entry conditions for this function are the same as the entry
        conditions of synchronize().

        It returns a list of (target path, Exception).

        This function blocks with impunity.
        '''
        # FIXME if adding params to the following call, add them above too
        will_sync, wont_sync, skipping, _ = self.compute_synchronization(
            unconditional=True
        )

        try:
            ensure_directories_exist([self.target_playlist_dir])
        except Exception as e:
            return [(self.target_playlist_dir, e)]

        excs = []
        written = []
        for p in self.playlists:
            newp = os.path.join(self.target_playlist_dir, os.path.basename(p))
            try:
                pdir = os.path.dirname(p)
                pf = open(p)
                pfl = pf.readlines()
                newpfl = []
                for l in pfl:
                    if not l.startswith("#"):
                        oldl = "# was: " + l
                        l = l.strip()
                        truel = os.path.abspath(os.path.join(pdir, l))
                        if truel in will_sync:
                            l = will_sync[truel]
                            l = os.path.relpath(l, self.target_playlist_dir)
                        elif truel in skipping:
                            l = skipping[truel]
                            l = os.path.relpath(l, self.target_playlist_dir)
                        elif truel in wont_sync:
                            l = "# not synced because of %s" % wont_sync[truel]
                        elif not l:
                            oldl = ""
                        else:
                            assert 0, (l, truel)
                        l = l + "\n"
                        l = oldl + l
                    newpfl.append(l)
                sync = True
                try:
                    with open(newp, "r") as oldpfl:
                        if oldpfl.read() == ''.join(newpfl):
                            sync = False
                except FileNotFoundError:
                    pass
                if sync:
                    if not dryrun:
                        with open(newp, "w") as newpf:
                            newpf.writelines(newpfl)
                            newpf.flush()
                        pf.close()
                    written.append(newp)
            except Exception as e:
                excs.append((newp, e))
        return written, excs

    def synchronize_deletions(self, deletor=os.unlink):
        '''
        Once synchronization of files and playlists has been done, sync
        of deletions is possible.

        The entry conditions for this function are the same as the entry
        conditions of synchronize_playlists().

        It returns a list of (target path, Exception).

        This function blocks with impunity.
        '''
        # FIXME if adding params to the following call, add them above too
        _, _, _, deleting = self.compute_synchronization(
            unconditional=True
        )

        excs = []
        for t in deleting:
            try:
                deletor(t)
            except Exception as e:
                excs.append((t, e))
        return excs

#=================== end synchronizer code ========================



#!/usr/bin/python3

from setuptools import setup
import os
import glob

dir = os.path.dirname(__file__)
path_to_main_file = os.path.join(dir, "lib/musictoolbox/__init__.py")
path_to_readme = os.path.join(dir, "README.md")
for line in open(path_to_main_file):
	if line.startswith('__version__'):
		version = line.split()[-1].strip("'").strip('"')
		break
else:
	raise ValueError('"__version__" not found in "lib/musictoolbox/__init__.py"')
readme = open(path_to_readme).read(-1)

classifiers = [
'Development Status :: 4 - Beta',
'Environment :: Console',
'Intended Audience :: End Users/Desktop',
'License :: OSI Approved :: GNU General Public License (GPL)',
'Operating System :: POSIX :: Linux',
'Programming Language :: Python :: 3 :: Only',
'Programming Language :: Python :: 3.7',
'Topic :: Utilities',
]

setup(
	name = 'musictoolbox',
	version=version,
	description = 'A set of utilities to help you groom your music collection',
	long_description = readme,
	author='Manuel Amador (Rudd-O)',
	author_email='rudd-o@rudd-o.com',
	license="GPL",
	url = 'http://github.com/Rudd-O/musictoolbox',
	package_dir=dict([
					("musictoolbox", "lib/musictoolbox"),
					]),
	classifiers = classifiers,
	packages = ["musictoolbox"],
	scripts = [ f for f in glob.glob(os.path.join("bin","*")) if not f.endswith("~") ],
	keywords = "mp3",
	requires = ["Twisted"],
	zip_safe=False,
)
